/*
 * sample.cc
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 *
 * Audio parts based on wmscope.c:
 *	Copyright (C) 1999 John Meacham
 *	released under the GPL version 2, see the file COPYING for details.
 *	wmscope release 3
 *	graphical sound display dockapp.
 *	john@foo.net
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 */

/* this is required for Solaris */
#define __EXTENSIONS__

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "sample.h"
#include "sample_msg.h"
#include "fft.h"

#define DEBUG		0
#define SHOW_TOP_FREQ	0
#include "debug.h"

/* for sound support either SOUND_OSS or SOUND_SOLARIS
 * must be defined. */
#if defined(SOUND_OSS)
#	if defined(__Linux__)
#		include <linux/soundcard.h>
#	elif defined(__FreeBSD__)
#		include <machine/soundcard.h>
#	else
#		include <sys/soundcard.h>
#	endif
#elif defined(SOUND_SOLARIS)
#	include <sys/audioio.h>
#else
#	error : No audio device setup.
#endif

#ifndef SAMPLING_SPEED
#	define SAMPLING_SPEED 44100
//#	define SAMPLING_SPEED 22050
#endif

#define SND_BUFFER_SIZE		FFT_BUFFER_SIZE			// in:  FFT_BUFFER_SIZE
#define FREQ_SIZE		(FFT_BUFFER_SIZE / 2 + 1)	// out: (FFT_BUFFER_SIZE / 2 + 1)


Sample::Sample(void) {
	DBG_MSG(1, "Sample::Sample: Initializing...");

	/* init vars. */
	finish = 0;
	pipe_fd[0] = 0;	/* read */
	pipe_fd[1] = 0;	/* write */
	
	/* create pipe */
	pipe(pipe_fd);

	/* init audio */
	open_audio();

	/* say bye */
	DBG_MSG(2, "Sample::Sample: Done.");
	return;	
}

Sample::~Sample(void) {
	DBG_MSG(1, "Sample::~Sample: Cleaning up...");

	/* set flag */
	finish = 1;
	
	/* close pipe */
	close(pipe_fd[1]);
	
	/* close audio */
	close_audio();
	
	DBG_MSG(2, "Sample::~Sample: Done.");
	return;
}

void *Sample::main(void) {
	sound_sample snd_buf[SND_BUFFER_SIZE];
	int bytes = 0, f = 0;
	fft_state *fft = NULL;
	float tmp_freq[FREQ_SIZE];
	int freq[FREQ_SIZE];
	//long int max_i;
	//float max_f;

	DBG_MSG(1, "Sample::main: Started.");
	DBG_MSG(8, "Sample::main: pipe: %d (r), %d (w)", this->pipe_fd[0], this->pipe_fd[1]);

	while (!finish) {
		/* create message */
		SampleMsg *msg = new SampleMsg(FREQ_SIZE);

		/* capture audio data */
		bzero(snd_buf, SND_BUFFER_SIZE);
		bytes = read_audio(snd_buf, SND_BUFFER_SIZE * sizeof(sound_sample));
		DBG_MSG(6, "Sample::main: read_audio: %d bytes.", bytes);

		/* urceni stredni hodnoty (aritm. prumer), posun na osu, abs a suma 
		 * -- zjisteni intenzity signalu pro detekci ticha */
		long sum = 0;
		for (int i = 0; i < SND_BUFFER_SIZE; i++) {
			sum += snd_buf[i];
		}
		int avg = sum / SND_BUFFER_SIZE;
		
		sum = 0;
		for (int i = 0; i < SND_BUFFER_SIZE; i++) {
			sum += abs((int)snd_buf[i] - avg);
		}
		msg->intensity = sum;
		DBG_MSG(6, "Sample::main: intensity = %Ld", msg->intensity);

		
		/* cast FFT spell */
		DBG_MSG(6, "Sample::main: FFT: init");
		fft = fft_init();
		if (!fft) {
			perror("fft_init");
			exit(EXIT_FAILURE);
		}
		DBG_MSG(6, "Sample::main: FFT: Casting spell...");
		fft_perform(snd_buf, tmp_freq, fft);
		/* tohle s tim pak dela xmms */
		for(int i = 0; i < FREQ_SIZE - 1; i++) {
			freq[i] = ((int)sqrt(tmp_freq[i + 1])) >> 8;
		}
		DBG_MSG(6, "Sample::main: FFT: Spell done.");
		fft_close(fft);

		/* copy data and freqs */
		memcpy(msg->freq, freq, sizeof(int) * FREQ_SIZE);
		for (f = 0; f < FREQ_SIZE; f ++) {
			msg->data[f].freq = f;
			msg->data[f].val = freq[f];
		}
		msg->sort_by_val();

#		if SHOW_TOP_FREQ == 1
#		define DIV	10	// 6
#		define FDIV	10	// 3
		char str[FREQ_SIZE / DIV];
		memset(str, ' ', FREQ_SIZE / DIV);
		char symbol[3] = { '*', ':', '.' };
		for (int i = 2; i >= 0; i--) {
			str[(msg->data[i].freq / FDIV) < FREQ_SIZE / DIV ? msg->data[i].freq / FDIV : FREQ_SIZE / DIV - 1]
				= symbol[i];
		}
		fprintf(stderr, " %4d(%3d) %4d(%3d) %4d(%3d) |%s|\n",
				msg->data[0].freq, msg->data[0].val,
				msg->data[1].freq, msg->data[1].val,
				msg->data[2].freq, msg->data[2].val,
				str);
#		endif
		
		/* send message to game */
		bytes = write(pipe_fd[1], &msg, sizeof(SampleMsg*));
		DBG_MSG(3, "Sample::main: Send %d bytes to pipe", bytes);
	}

	close(pipe_fd[1]);
	
	DBG_MSG(1, "Sample::main: Exit.");
	return NULL;
}

int Sample::get_pipe(void) {
	return pipe_fd[0];
}

int Sample::open_audio(char *audio_device) {
#ifdef SOUND_OSS
	int arg = 0x00020008;
	int speed = SAMPLING_SPEED;
	int format = AFMT_S16_LE;

	DBG_MSG(1, "Sample::open_audio: Initializing OSS \"%s\"...", audio_device);
	
	audio_fd = open(audio_device, O_RDONLY);
	if(audio_fd == -1) {
		perror(audio_device);
		return 0;
	}
	if (ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &arg)) {
		perror(audio_device);
		return 0;
	}
	ioctl(audio_fd, SNDCTL_DSP_SPEED, &speed);
	//fprintf(stderr, "Audio: Sampling Rate: %i\n", speed);
	if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &format) == -1) {
		perror("SNDCTL_DSP_SETFMT");
		return 0;
	}
	if(format != AFMT_S16_LE)
		fprintf(stderr,
			"Audio: Could not set %s to signed 16 bit mode\n"
			"\tattempting to ignore error.\n", audio_device);
	/*
	else 
		fprintf(stderr, "Audio: Mode: 16 bit signed.\n");
	*/

	DBG_MSG(2, "Sample::open_audio: done");
	return 1;

#elif defined(SOUND_SOLARIS)

	DBG_MSG(1, "Sample::open_audio: Initializing Solaris audio...");
	
	audio_info_t info;
	audio_fd = open(audio_device, O_RDONLY);
	if(audio_fd == -1) { 
		perror(audio_device);
		sound_on = false;
		return 0;
	}
	
	ioctl(audio_fd, AUDIO_GETINFO, &info);
	info.play.sample_rate = info.record.sample_rate = SAMPLING_SPEED;
	info.play.channels = info.record.channels = 1;
	info.play.precision = info.record.precision = 16;
	info.play.encoding = info.record.encoding = AUDIO_ENCODING_LINEAR;
	ioctl(audio_fd, AUDIO_SETINFO, &info);
	ioctl(audio_fd, AUDIO_GETINFO, &info);
	
	/*
	printf("Audio: speed %i channels %i prec %i enc %i\n",
	       info.record.sample_rate, info.record.channels,
	       info.record.precision, info.record.encoding);
	*/
	DBG_MSG(2, "Sample::open_audio: done");
	return 1;
#endif
}


int Sample::read_audio(sound_sample *buf, int buf_size) {
	int i = 0, bytes = 1;
	DBG_MSG(2, "Sample::read_audio: Reading...");
	while(i < buf_size && bytes > 0) {
		bytes = read(audio_fd, buf + i , buf_size - i);
		i += bytes;
		DBG_MSG(2, "Sample::read_audio: Read %d bytes", bytes);
	}
	DBG_MSG(2, "Sample::read_audio: Done, totaly %d bytes read", i);
	return i;
}

int Sample::close_audio(void) {
	DBG_MSG(1, "Sample::close_audio: Closing audio (audio_fd = %d)...", audio_fd);
	if(close(audio_fd)) {
		perror("close audio");
		DBG_MSG(1, "Sample::close_audio: error");
		return 0;
	}
	DBG_MSG(2, "Sample::close_audio: done");
	return 1;
	
}

int Sample::scan_freq(char *buf) {
	return 0;
}

