/*
 * screen.cc
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ncurses.h>
#include <signal.h>

#include "screen.h"
#include "game_world.h"
#include "game_object.h"
#include "sample_msg.h"

#define DEBUG 0
#include "debug.h"

#define ATTR_ST			(COLOR_PAIR(1))
#define ATTR_ST_LIGHT		(COLOR_PAIR(2) | A_BOLD)
#define ATTR_ST_DARK		(COLOR_PAIR(3) | A_BOLD)
#define ATTR_BORDER		(COLOR_PAIR(1))
#define ATTR_FREQ		(COLOR_PAIR(4))
#define ATTR_FREQ_H		(COLOR_PAIR(5) | A_BOLD)
#define ATTR_FREQ_RULE		(COLOR_PAIR(6))
#define ATTR_FREQ_RULE_INT	(COLOR_PAIR(7) | A_BOLD)
#define ATTR_FREQ_RULE_INT_H	(COLOR_PAIR(8) | A_BOLD)

/*****************************************************************************
 * Inicializace obrazovky
 */
Screen::Screen(void) {
	DBG_MSG(1, "Screen::Screen: Initializing...");


	/* ncurses init */
	initscr();
	noecho();
	cbreak();
	crmode();
	keypad(stdscr, 1);
	if (has_colors() != ERR) {
		DBG_MSG(6, "Screen::Screen: Initializing colors...");
		start_color();
		init_pair(1, COLOR_WHITE,	COLOR_BLACK);
		init_pair(2, COLOR_YELLOW,	COLOR_BLACK);
		init_pair(3, COLOR_BLACK,	COLOR_BLACK);
		init_pair(4, COLOR_CYAN,	COLOR_BLACK);
		init_pair(5, COLOR_WHITE,	COLOR_BLACK);
		init_pair(6, COLOR_WHITE,	COLOR_BLACK);
		init_pair(7, COLOR_WHITE,	COLOR_BLUE);
		init_pair(8, COLOR_WHITE,	COLOR_GREEN);

		/* barvy pro objekty */
		init_pair(10, COLOR_BLACK,	COLOR_BLACK);
		init_pair(11, COLOR_BLUE,	COLOR_BLACK);
		init_pair(12, COLOR_GREEN,	COLOR_BLACK);
		init_pair(13, COLOR_RED,	COLOR_BLACK);
		init_pair(14, COLOR_CYAN,	COLOR_BLACK);
		init_pair(15, COLOR_MAGENTA,	COLOR_BLACK);
		init_pair(16, COLOR_YELLOW,	COLOR_BLACK);
		init_pair(17, COLOR_BLACK,	COLOR_BLACK);
		init_pair(18, COLOR_BLACK,	COLOR_BLACK); // + A_BOLD
		init_pair(19, COLOR_BLUE,	COLOR_BLACK); //    :
		init_pair(20, COLOR_GREEN,	COLOR_BLACK); //    .
		init_pair(21, COLOR_RED,	COLOR_BLACK);
		init_pair(22, COLOR_CYAN,	COLOR_BLACK);
		init_pair(23, COLOR_MAGENTA,	COLOR_BLACK);
		init_pair(24, COLOR_YELLOW,	COLOR_BLACK);
		init_pair(25, COLOR_WHITE,	COLOR_BLACK);
	} else {
		fprintf(stderr, "Cannot initialize colors!\n");
	}
	
	resized = 0;
	draw_all();

	DBG_MSG(2, "Screen::Screen: Done.");
}


/*****************************************************************************
 * Vyitn obrazovky
 */
Screen::~Screen(void) {
	DBG_MSG(1, "Screen::~Screen: Cleaning up...");
	
	/* ncurses */
	endwin();

	DBG_MSG(2, "Screen::~Screen: Done.");
}


/*****************************************************************************
 * Vykresleni celeho sveta -- neumi to scrollovani, ale aspon to vycentruje
 * svet na obrazovce
 */
void Screen::render_world(GameWorld *world) {
	DBG_MSG(2, "Screen::render_world");
	GObject *object = NULL;
	GSymbol *new_symbol = NULL;

	world_height = world->get_height();
	world_width = world->get_width();

	DBG_MSG(2, "Screen::render_world: Rendering world (width = %d, height = %d)...", world_width, world_height);

	
	for (std::vector<GSymbol *>::iterator i = symbol.begin(); i != symbol.end(); i++)
		delete *i;
	symbol.clear();
	symbol.resize(world_height * world_width, NULL);

	world->get_child_rewind();
	while ((object = world->get_child())) {
		new_symbol = object->draw();
		DBG_MSG(9, "Screen::render_world: Found object (symbol is '%c').", new_symbol->symbol);
		if (new_symbol->x >= 0 && new_symbol->x < world_width
				&& new_symbol->y >= 0 && new_symbol->y < world_height
				&& new_symbol->z >= 0 && new_symbol->z <= 2)
		{
			DBG_MSG(9, "Screen::render_world: Object is OK.");
			if (symbol[new_symbol->x + new_symbol->y * world_width]
					&& symbol[new_symbol->x + new_symbol->y * world_width]->z < new_symbol->z)
			{
				delete symbol[new_symbol->x + new_symbol->y * world_width];
				symbol[new_symbol->x + new_symbol->y * world_width] = NULL;
			}

			if (symbol[new_symbol->x + new_symbol->y * world_width] == NULL) {
				symbol[new_symbol->x + new_symbol->y * world_width] = new_symbol;
				DBG_MSG(9, "Screen::render_world: Object stored (symbol is '%c').",
						symbol[new_symbol->x + new_symbol->y * world_width]->symbol);
			}
		}
	}

	DBG_MSG(2, "Screen::render_world: Done.");
}


/*****************************************************************************
 * Renderovani frekvencniho spektra do pole intu podle velikosti obrazovky
 *
 * Prevede pole floatu o velke sirce na pole intu o sirce obrazovky.
 *
 * Pri zmene velikosti obrazovky spektrum zmizi a vykresli se az po dalsim
 * renderovani
 *
 * prepocet:
 * 	1. sloucit frekvence do sloupcu (pokud je obrazovka sirsi nez spektrum,
 *	   tak se jen kopiruje)
 *  	   - zaklad: 1 sloupec grafu = >1 sloupec spektra
 *  	   
 * 	2. upravit vysku sloupcu
 * 	   - najit maximum == nejvyssi sloupec
 * 	   - prepocitat screen_height - world_height - 5 == nejvyssi sloupec
 *
 * max intensity = 35'000'000
 */
void Screen::render_freq(int *freq, int size, int min_colorized, int intensity, int intensity_min) {
	int do_zoom = screen_width - 2 < size;
	int bar_size = do_zoom ? screen_width - 2 : size;
	int bar[bar_size];
	int max = 0;
	DBG_MSG(2, "Screen::render_freq: size = %d, min_colorized = %f; do_zoom = %d",
			size, min_colorized, do_zoom);
	
	bzero(bar, bar_size * sizeof(int));

	
	/* slouceni frekvenci */
	if (do_zoom) {
		int f = 0, b = 0;
		int f_step = (int)((float)size / (float)bar_size);
		/* kaslu na zbytecek na konci */
		while (b < bar_size && f < size) {
			if (f % f_step == f_step - 1) {
				max = bar[b] > max ? bar[b] : max;
				b++;
			}
			bar[b] += freq[f++];
		}
		hscale = (float)b / (float)f;
	} else /* screen_width >= size */ {
		for (int i = 0; i < size; i++) {
			bar[i] = freq[i];
			max = bar[i] > max ? bar[i] : max;
		}
		memcpy(bar, freq, sizeof(int) * size);
		hscale = 1.0;
	}

	/* uprava vysek */
	bar_height = screen_height - world_height - 5;
	bar_width = bar_size;
	freq_bar.resize(bar_size);
	float mul = (float)bar_height / (float)max;
	for (int i = 0; i < bar_size; i++)
		freq_bar[i] = (int)(bar[i] * mul);
	freq_bar_max = max;
	freq_bar_min_colorized = (int)(mul * min_colorized);

	/* uprava intensity */
	this->intensity = (int)(((float)intensity / (float)(intensity_min * 8)) * bar_size);
	this->intensity_min = (int)(((float)intensity_min / (float)(intensity_min * 8)) * bar_size);
	
	DBG_MSG(4, "Screen::render_freq: max = %d, min_colorized = %d",
			freq_bar_max, freq_bar_min_colorized);
}

/*****************************************************************************
 * Prida frekvenci na pravitko s frekvencema
 */
void Screen::add_freq(float from, float to, char label) {
	s_freq_label item;
	item.from = from;
	item.to = to;
	item.label = label;
	freq_rule.push_back(item);
}
void Screen::clear_freq_rule(void) {
	freq_rule.clear();
}

/*****************************************************************************
 * Prekresleni obrazovky
 */
void Screen::redraw(void) {
	DBG_MSG(2, "Screen::redraw: Redrawing screen...");
	endwin();
	refresh();
	draw_all();
	touchwin(stdscr);
	refresh();
	DBG_MSG(2, "Screen::redraw: Redrawing screen... Done.");
}


/*****************************************************************************
 * Ulozeni flagu, ze se ma obrazovka refreshnout dukladne
 */
void Screen::resize_notify(void) {
	resized = 1;
}

/*****************************************************************************
 * main() threadu
 */
/*
void *Screen::main(void) {
	DBG_MSG(1, "Screen::main: Started.");
	sleep(4);
	DBG_MSG(1, "Screen::main: Exit.");
	return NULL;
}
*/


/*****************************************************************************
 * Vykresli vsechno
 */
void Screen::draw_all(void) {
	if (resized) {
		endwin();
		refresh();
		resized = 0;
	}
	erase();
	getmaxyx(stdscr, screen_height, screen_width);
	draw_status();
	draw_world();
	draw_freq();
	flush();
}



/*****************************************************************************
 * Vykresleni hraciho sveta
 */
void Screen::draw_world(void) {
	if (resized) {
		draw_all();
		return;
	}

	DBG_MSG(2, "Screen::draw_world: Drawing world (screen_width = %d, screen_height = %d)...",
			screen_width, screen_height);
	
	//int voff = ((screen_height - world_height - 2) / 2) + 2;
	int voff = 3;
	int hoff = ((screen_width - world_width) / 2);

	/* ramecek */
	attrset(ATTR_BORDER);
	mvaddch(voff - 1, hoff - 1, ACS_ULCORNER);
	mvaddch(voff - 1, hoff + world_width, ACS_URCORNER);
	mvaddch(voff + world_height, hoff - 1, ACS_LLCORNER);
	mvaddch(voff + world_height, hoff + world_width, ACS_LRCORNER);
	
	for (int col = 0; col < world_width; col++)
		mvaddch(voff - 1, hoff + col, ACS_HLINE);
	
	for (int col = 0; col < world_width; col++)
		mvaddch(voff + world_height, hoff + col, ACS_HLINE);

	for (int row = 0; row < world_height; row++)
		mvaddch(voff + row, hoff - 1, ACS_VLINE);
	
	for (int row = 0; row < world_height; row++)
		mvaddch(voff + row, hoff + world_width, ACS_VLINE);

	/* svet */
	for (int row = 0; row < world_height; row++) {
		for (int col = 0; col < world_width; col++) {
			if (symbol[row * world_width + col])
				attrset(COLOR_PAIR(symbol[row * world_width + col]->color + 10) 
						| (symbol[row * world_width + col]->color > 7 ? A_BOLD : 0));
			else
				attrset(COLOR_PAIR(17));
				
			mvaddch(voff + row, hoff + col,
					symbol[row * world_width + col] ?  symbol[row * world_width + col]->symbol : ' ');
		}
	}
	
	DBG_MSG(2, "Screen::draw_world: Done.");
}


/*****************************************************************************
 * Vykresleni cudliku
 */
void Screen::draw_status(void) {
	if (resized) {
		draw_all();
		return;
	}

	DBG_MSG(6, "Screen::draw_status: drawing...");

	/*zahradnik */
	move(0, 2);
	attrset(ATTR_ST);
	printw("Zahradnik: ");
	
	attrset(status[2] & 2 ? ATTR_ST_LIGHT : ATTR_ST_DARK);	
	printw(" <-");
	
	attrset(ATTR_ST);
	printw(" %4d ", status[3]);
	
	attrset(status[2] & 1 ? ATTR_ST_LIGHT : ATTR_ST_DARK);	
	printw(" ->");

	/* Krtek */
	move(0, screen_width - 20);
	attrset(ATTR_ST);
	printw("Krtek: ");
	
	attrset(status[0] & 2 ? ATTR_ST_LIGHT : ATTR_ST_DARK);	
	printw(" <-");
	
	attrset(ATTR_ST);
	printw(" %4d ", status[1]);
	
	attrset(status[0] & 1 ? ATTR_ST_LIGHT : ATTR_ST_DARK);	
	printw(" ->");

	move(1, 0);
	attrset(ATTR_ST);
	for (int i = 0; i < screen_width; i++)
		insch(ACS_HLINE);

}


/*****************************************************************************
 * Vykresleni frekv. spektra
 */
void Screen::draw_freq(void) {
	
	DBG_MSG(6, "Screen::draw_freq: drawing...");
	if (resized) {
		draw_all();
		return;
	}

	int voff = screen_height - bar_height;
	unsigned int i;

	if (resized)
		draw_all();

	if (voff >= screen_height)
		return;
	
	for (i = 0; i < freq_bar.size(); i++) {
		attrset(freq_bar[i] > freq_bar_min_colorized ? ATTR_FREQ_H : ATTR_FREQ);
		for (int r = 0; r < bar_height - 1; r++) {
			mvaddch(voff + r, i + 1, freq_bar[i] > bar_height - r ? '#' : ' ');
		}
	}
	
	/* pravitko */
	attrset(ATTR_FREQ_RULE_INT);
	for (i = 0; i < freq_bar.size() - 1; i++) {
		if (i == (unsigned)intensity_min) {
			attrset(ATTR_FREQ_RULE_INT_H);
		}
		if (i > (unsigned)intensity) {
			attrset(ATTR_FREQ_RULE);
		}
		for (FreqRule::iterator b = freq_rule.begin(); b != freq_rule.end(); b++) {
			if (i > b->from * hscale && i < b->to * hscale) {
				mvaddch(screen_height - 1, i + 2, b->label);
				break;
			} else {
				mvaddch(screen_height - 1, i + 2, ' ');
			}
		}
	}
	attrset(ATTR_FREQ_RULE);
	mvaddch(screen_height - 1, 1, ACS_VLINE);
	mvaddch(screen_height - 1, freq_bar.size(), ACS_VLINE);
	DBG_MSG(6, "Screen::draw_freq: done");
}


/*****************************************************************************
 * Nastaven veci ve statusbaru
 */
void Screen::set_status(int krtek_key, int krtek_score, int zahradnik_key, int zahradnik_score) {
	status[0] = krtek_key & 3;
	status[1] = krtek_score;
	status[2] = zahradnik_key & 3;
	status[3] = zahradnik_score;
	draw_status();
}


/*****************************************************************************
 * Jen opravdu vykresleni toho co je pripraveno
 */
void Screen::flush(void) {
	if (resized)
		draw_all();
	move(screen_height - 1, 0);
	refresh();
}
